/*
 *  ImgView  A simple image viewer
 *
 *  Copyright (C) 2013 Mats Peterson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  Please send any comments/bug reports to
 *  matsp888@yahoo.com  (Mats Peterson)
 */

#include <xsimple.h>
#include <stdio.h>
#include <stdlib.h>
#include <wchar.h>
#include <string.h>
#include <gif_lib.h>
#include "common.h"
#include "../imgview.h"

static int tindex;

static char *errstr[] = {
    "Failed to open given file",
    "Failed to Read from given file",
    "Given file is NOT GIF file",
    "No Screen Descriptor detected",
    "No Image Descriptor detected",
    "Neither Global Nor Local color map",
    "Wrong record type detected",
    "#Pixels bigger than Width * Height",
    "Failed to allocate required memory",
    "Failed to close given file",
    "Given file was not opened for read",
    "Image is defective, decoding aborted",
    "Image EOF detected, before image complete"
};


static void errmsg(void)
{
    char *s;
    int err;

    err = GifLastError();
    s = (err >= D_GIF_ERR_OPEN_FAILED && err <= D_GIF_ERR_EOF_TOO_SOON) ?
            errstr[err - D_GIF_ERR_OPEN_FAILED] : "Unknown error";
    message("GIF Library Error", s, NULL);
}

XImage *getimage(GifFileType *gf)
{
    XImage *img = NULL;
    int w, h;
    ColorMapObject *cmap;
    int bilvl = 0;
    unsigned char *buf = NULL;
    GifColorType *grgb;
    RGB bg, rgb;
    int x, y, bit;
    unsigned char *rgbabuf = NULL;

    if (DGifGetImageDesc(gf) == GIF_ERROR) {
        errmsg(); goto end;
    }

    w = gf->Image.Width; h = gf->Image.Height;

    cmap = (gf->SColorMap ? gf->SColorMap : gf->Image.ColorMap);

    {
        char par_str[80] = "";
        if (gf->GifVersion == 89 && gf->SPixelAspectRatio) {
            double par = ((double)(gf->SPixelAspectRatio) + 15.) / 64.;
            sprintf(par_str, "Pixel Aspect Ratio (w/h): %g\n", par);
            v.par = 1. / par;
        }
        sprintf(v.xinfostr,
                "%s"
                "Color Resolution: %d bits\n"
                "Bits Per Pixel: %d\n"
                "Interlace: %s\n"
                "Version: %da",
                par_str,
                gf->SColorResolution, cmap->BitsPerPixel,
                gf->Image.Interlace ? "Yes" : "No",
                gf->GifVersion);
    }

    if (tindex == -1 && cmap->ColorCount == 2) {
        GifColorType rgb0 = cmap->Colors[0], rgb1 = cmap->Colors[1];
        v.black = v.white = -1;
        if (rgb0.Red == 0 && rgb0.Green == 0 && rgb0.Blue == 0)
            v.black = 0;
        else if (rgb0.Red == 255 && rgb0.Green == 255 && rgb0.Blue == 255)
            v.white = 0;
        if (rgb1.Red == 0 && rgb1.Green == 0 && rgb1.Blue == 0)
            v.black = 1;
        else if (rgb1.Red == 255 && rgb1.Green == 255 && rgb1.Blue == 255)
            v.white = 1;
        if (v.black != -1 && v.white != -1) bilvl++;
    }

    if (! bilvl && v.hOutProfile) {
        if (v.iprof)
            v.hInProfile = cmsOpenProfileFromMem(v.iprof, v.iproflen);
        else
            v.hInProfile = v.hRGBProfile;
        if (v.hInProfile && !isoutprof()) {
            v.hTransform = cmsCreateTransform(
                    v.hInProfile,
                    TYPE_RGBA_8,
                    v.hOutProfile,
                    TYPE_RGBA_8,
                    INTENT_PERCEPTUAL, 0);
            if (v.hTransform) {
                rgbabuf = (unsigned char *)malloc(w * 4);
                bg.r = v.tbgcolor_rgb.red >> 8;
                bg.g = v.tbgcolor_rgb.green >> 8;
                bg.b = v.tbgcolor_rgb.blue >> 8;
            }
        }
    }

    v.pad = 0;
    img = mkximage(w, h, bilvl ? 1 : xv.depth, &v.pad);

    if (bilvl) memset(img->data, 0, img->bytes_per_line * h);

    if (gf->Image.Interlace) {
        int start[4] = {0, 4, 2, 1};
        int off[4] = {8, 8, 4, 2};
        int grp, i;

        buf = (unsigned char *)malloc(w * h);
        if (DGifGetLine(gf, buf, w * h) == GIF_ERROR) {
#ifndef GIF_NICE
            errmsg(); rmximage(img); img = NULL; goto end;
            goto end;
#endif
        }

        for (grp = 0, i = 0; grp < 4; grp++) {
            for (y = start[grp]; y < h; y += off[grp], i++) {
                dp = (unsigned char *)img->data + y * img->bytes_per_line;
                idp = (unsigned int *)dp;
                if (bilvl) {
                    for (x = 0, bit = 7; x < w; x++, bit--) {
                        if (bit < 0) { dp++; bit = 7; }
                        *dp |= buf[i * w + x] << bit;
                    }
                } else {
                    if (v.hTransform) {
                        unsigned char *p1 = buf + i * w;
                        unsigned char *p2 = rgbabuf;
                        for (x = 0; x < w; x++) {
                            int index = *p1++;
                            grgb = &(cmap->Colors[index]);
                            *p2++ = grgb->Red;
                            *p2++ = grgb->Green;
                            *p2++ = grgb->Blue;
                            *p2++ = (tindex != -1 && index == tindex) ?
                                    0 : 255;
                        }
                        cmsDoTransform(v.hTransform, rgbabuf, rgbabuf, w);
                        p1 = rgbabuf;
                        for (x = 0; x < w; x++) {
                            unsigned char a;
                            rgb.r = *p1++; rgb.g = *p1++; rgb.b = *p1++;
                            a = *p1++;
                            if (a == 0) rgb = bg;
                            DOPIXEL(rgb.r, rgb.g, rgb.b, 255);
                        }
                    } else {
                        for (x = 0; x < w; x++) {
                            int index = buf[i * w + x];
                            if (tindex != -1 && index == tindex)
                                DOPIXEL2(v.tbgcolor);
                            else {
                                grgb = &(cmap->Colors[index]);
                                DOPIXEL(grgb->Red, grgb->Green,
                                        grgb->Blue, 255);
                            }
                        }
                    }
                }
                if (v.pad) dp += v.pad;
            }
        }
    } else {
        dp = (unsigned char *)img->data;
        idp = (unsigned int *)dp;

        buf = (unsigned char *)malloc(w);

        for (y = 0; y < h; y++) {
            if (DGifGetLine(gf, buf, w) == GIF_ERROR) {
#ifndef GIF_NICE
                errmsg(); rmximage(img); img = NULL;
#endif
                goto end;
            }
            if (bilvl) {
                for (x = 0, bit = 7; x < w; x++, bit--) {
                    if (bit < 0) { dp++; bit = 7; }
                    *dp |= buf[x] << bit;
                }
                dp++;
            } else {
                if (v.hTransform) {
                    unsigned char *p1 = buf;
                    unsigned char *p2 = rgbabuf;
                    for (x = 0; x < w; x++) {
                        int index = *p1++;
                        grgb = &(cmap->Colors[index]);
                        *p2++ = grgb->Red;
                        *p2++ = grgb->Green;
                        *p2++ = grgb->Blue;
                        *p2++ = (tindex != -1 && index == tindex) ? 0 :
                                255;
                    }
                    cmsDoTransform(v.hTransform, rgbabuf, rgbabuf, w);
                    p1 = rgbabuf;
                    for (x = 0; x < w; x++) {
                        unsigned char a;
                        rgb.r = *p1++; rgb.g = *p1++; rgb.b = *p1++;
                        a = *p1++;
                        if (a == 0) rgb = bg;
                        DOPIXEL(rgb.r, rgb.g, rgb.b, 255);
                    }
                } else {
                    for (x = 0; x < w; x++) {
                        int index = buf[x];
                        if (tindex != -1 && index == tindex)
                            DOPIXEL2(v.tbgcolor);
                        else {
                            grgb = &(cmap->Colors[index]);
                            DOPIXEL(grgb->Red, grgb->Green, grgb->Blue,
                                    255);
                        }
                    }
                }
            }
            if (v.pad) dp += v.pad;
        }
    }

end:
    if (v.hTransform) {
        cmsDeleteTransform(v.hTransform); v.hTransform = NULL;
        if (rgbabuf) free(rgbabuf);
    }
    if (v.hInProfile) {
        if (v.iprof) cmsCloseProfile(v.hInProfile); v.hInProfile = NULL;
    }
    if (v.iprof) free(v.iprof); v.iprof = NULL;
    if (buf) free(buf);
    return img;
}

static int getextension(GifFileType *gf)
{
    int code;
    GifByteType *ext;
    int appbgn = 0, profext = 0;

    if (DGifGetExtension(gf, &code, &ext) == GIF_ERROR) {
#ifndef GIF_NICE
        errmsg();
#endif
        return 0;
    }
    while (ext) {
        switch (code) {
            case GRAPHICS_EXT_FUNC_CODE:
                if (ext[1] & 1) tindex = ext[4];
                break;
            case COMMENT_EXT_FUNC_CODE:
                {
                char *s;
                if (v.ncomments == MAX_COMMENTS) break;
                v.comments[v.ncomments].key = strdup("");
                s = (char *)malloc(ext[0] + 1);
                strncpy(s, (const char *)(&ext[1]), ext[0]);
                *(s + ext[0]) = '\0';
                if (isutf8(s))
                    v.comments[v.ncomments].text = strdup(s);
                else
                    v.comments[v.ncomments].text = strdup(win2utf8(s));
                free(s);
                v.ncomments++;
                }
                break;
            case APPLICATION_EXT_FUNC_CODE:
                if (! appbgn) {
                    profext = !memcmp(&ext[1], "ICCRGBG1012", 11);
                    appbgn++;
                } else if (! v.iprof && profext) {
                    int bs = ext[0];
                    v.iprof = (char *)malloc(bs);
                    memcpy(v.iprof, &ext[1], bs);
                    v.iproflen = bs;
                } else if (v.iprof) {
                    int bs = ext[0];
                    v.iprof = (char *)realloc(v.iprof, v.iproflen + bs);
                    memcpy(v.iprof + v.iproflen, &ext[1], bs);
                    v.iproflen += bs;
                }
        }
        if (DGifGetExtensionNext(gf, &ext) == GIF_ERROR) {
#ifndef GIF_NICE
            errmsg();
#endif
            return 0;
        }
    }
    return 1;
}

XImage *readgif(int fd)
{
    GifFileType *gf;
    unsigned char *buf = NULL;
    XImage *img = NULL;

    if (! (gf = DGifOpenFileHandle(fd))) {
        errmsg(); goto end;
    }

    tindex = -1;

    for (;;) {
        GifRecordType type;
        if (DGifGetRecordType(gf, &type) == GIF_ERROR) {
#ifndef GIF_NICE
            errmsg();
#endif
            goto end;
        }
        switch (type) {
            case IMAGE_DESC_RECORD_TYPE:
                if (! img) {
                    if (! (img = getimage(gf))) goto end;
                } else {
                    int w, h;
                    if (DGifGetImageDesc(gf) == GIF_ERROR) {
#ifndef GIF_NICE
                        errmsg();
#endif
                        goto end;
                    }
                    w = gf->Image.Width; h = gf->Image.Height;
                    buf = (unsigned char *)malloc(w * h);
                    if (DGifGetLine(gf, buf, w * h) == GIF_ERROR) {
#ifndef GIF_NICE
                        errmsg();
#endif
                        goto end;
                    }
                    free(buf); buf = NULL;
                }
                break;
            case EXTENSION_RECORD_TYPE:
                if (! getextension(gf)) goto end;
                break;
            case TERMINATE_RECORD_TYPE:
                goto end;
            default:
                break;
        }
    }

end:
    if (buf) free(buf);
    DGifCloseFile(gf);
    return img;
}
