/*
 *  ImgView  A simple image viewer
 *
 *  Copyright (C) 2013 Mats Peterson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  Please send any comments/bug reports to
 *  matsp888@yahoo.com  (Mats Peterson)
 */

#include <xsimple.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <wchar.h>
#include <jpeglib.h>
#include <setjmp.h>
#include <strutil.h>
#include "common.h"
#include "../imgview.h"

static int bigend;

typedef unsigned short uint16;
typedef unsigned int uint32;


static uint16 getshort(uint16 n)
{
    return bigend ? ((n >> 8) & 0xff) | ((n << 8) & 0xff00) : n;
}

static uint32 getlong(uint32 n)
{
    return bigend ? ((n >> 24) & 0xff) | ((n >> 8) & 0xff00) |
            ((n << 8) & 0xff0000) | ((n << 24) & 0xff000000) : n;
}

static void getexifres(char *hdr, double *xres, double *yres,
        uint16 *unit)
{
    uint32 ifdoffs;
    uint16 nflds;
    char *p;
    int i, done = 0;

    bigend = *hdr == 'M';
    ifdoffs = getlong(*(uint32 *)(hdr + 4));
    p = hdr + ifdoffs;
    nflds = getshort(*(uint16 *)p); p += 2;

    for (i = 0; i < nflds; i++, p += 12) {
        uint16 tagid = getshort(*(uint16 *)p);
        uint32 voffs, num, den;
        double *p2;
        switch (tagid) {
            case 0x11a:
            case 0x11b:
                voffs = getlong(*(uint32 *)(p + 8));
                num = getlong(*(uint32 *)(hdr + voffs));
                den = getlong(*(uint32 *)(hdr + voffs + 4));
                p2 = (tagid == 0x11a) ? xres : yres;
                *p2 = (double)num / (double)den;
                done++;
                break;
            case 0x128:
                *unit = getshort(*(uint16 *)(p + 8));
                done++;
        }
        if (done == 3) break;
    }
}

struct my_error_mgr {
    struct jpeg_error_mgr pub;
    jmp_buf setjmp_buffer;
};

typedef struct my_error_mgr *my_error_ptr;

METHODDEF(void) my_error_exit(j_common_ptr cinfo)
{
    char s[1024];
    my_error_ptr myerr = (my_error_ptr)cinfo->err;
    cinfo->err->format_message(cinfo, s);
    message("JPEG Library Error", s, NULL);
    longjmp(myerr->setjmp_buffer, 1);
}

XImage *readjpeg(int fd, int *jpgfmt)
{
    FILE *f;
    struct jpeg_decompress_struct cinfo;
    struct my_error_mgr jerr;
    jpeg_saved_marker_ptr mp;
    jpeg_saved_marker_ptr profmp[256];
    XImage *volatile img = NULL;
    JSAMPARRAY buf;
    JSAMPLE *bp;
    int row_stride;
    int w, h;
    RGB rgb, *rgbp;
    int x, y;
    int i;
    JSAMPLE *g2rbuf = NULL;

    f = fdopen(fd, "r");

    cinfo.err = jpeg_std_error(&jerr.pub);
    jerr.pub.error_exit = my_error_exit;
    if (setjmp(jerr.setjmp_buffer)) {
        if (v.hTransform) {
            cmsDeleteTransform(v.hTransform); v.hTransform = NULL;
            if (g2rbuf) free(g2rbuf);
        }
        if (v.hInProfile) {
            if (v.iprof) cmsCloseProfile(v.hInProfile); v.hInProfile = NULL;
        }
        if (v.iprof) free(v.iprof);
        jpeg_destroy_decompress(&cinfo);
        fclose(f);
        if (img) rmximage(img);
        return NULL;
    }

    jpeg_create_decompress(&cinfo);
    jpeg_stdio_src(&cinfo, f);
    (void)jpeg_save_markers(&cinfo, JPEG_APP0+1, 0xffff);
    if (v.hOutProfile)
        (void)jpeg_save_markers(&cinfo, JPEG_APP0+2, 0xffff);
    (void)jpeg_save_markers(&cinfo, JPEG_COM, 0xffff);
    (void)jpeg_read_header(&cinfo, TRUE);
    jpeg_calc_output_dimensions(&cinfo);

    *jpgfmt = FMT_JPEG_JIF;

    for (i = 0; i < 256; i++) profmp[i] = NULL;

    for (mp = cinfo.marker_list; mp; mp = mp->next) {
        if (mp->marker == JPEG_COM) {
            char *s;
            if (v.ncomments == MAX_COMMENTS) continue;
            v.comments[v.ncomments].key = strdup("");
            s = (char *)malloc(mp->data_length + 1);
            strncpy(s, (const char *)mp->data, mp->data_length);
            *(s + mp->data_length) = '\0';
            if (isutf8(s))
                v.comments[v.ncomments].text = strdup(s);
            else
                v.comments[v.ncomments].text = strdup(win2utf8(s));
            free(s);
            v.ncomments++;
        } else if (mp->marker == JPEG_APP0+1 &&
                    !strcmp((const char *)mp->data, "Exif")) {
            double xres = 72., yres = 72.;
            uint16 unit = 2;
            getexifres((char *)mp->data + 6, &xres, &yres, &unit);
            if (unit == 2) {
                v.ippi[0] = xres;
                v.ippi[1] = yres;
            } else {
                v.ippi[0] = xres * 2.54;
                v.ippi[1] = yres * 2.54;
            }
            if (xres && yres) v.par = xres / yres;
            sprintf(v.xinfostr,
                    "Resolution: %g × %g pixels per inch",
                    v.ippi[0], v.ippi[1]);
            *jpgfmt = FMT_JPEG_EXIF;
        } else if (v.hOutProfile && mp->marker == JPEG_APP0+2 &&
                    !strcmp((const char *)mp->data, "ICC_PROFILE")) {
            unsigned char cn = *((unsigned char *)mp->data + 12);
            profmp[cn] = mp;
            v.iproflen += mp->data_length - 14;
        }
    }

    if (v.hOutProfile) {
        int nc = cinfo.output_components;
        if (v.iproflen) {
            v.iprof = (char *)malloc(v.iproflen);
            v.iproflen = 0;
            for (i = 0; i < 256; i++) {
                jpeg_saved_marker_ptr mp = profmp[i];
                if (mp) {
                    int len = mp->data_length - 14;
                    memcpy(v.iprof + v.iproflen, mp->data + 14, len);
                    v.iproflen += len;
                }
            }
            v.hInProfile = cmsOpenProfileFromMem(v.iprof, v.iproflen);
        } else {
            v.hInProfile = nc == 4 ? v.hCMYKProfile :
                    nc == 3 ? v.hRGBProfile : v.hGrayProfile;
        }
        if (v.hInProfile && !isoutprof()) {
            v.hTransform = cmsCreateTransform(
                    v.hInProfile,
                    nc == 4 ? TYPE_CMYK_8_REV : nc == 3 ? TYPE_RGB_8 :
                            TYPE_GRAY_8,
                    v.hOutProfile,
                    TYPE_RGB_8,
                    INTENT_PERCEPTUAL, 0);
            if (v.hTransform && cinfo.output_components == 1)
                g2rbuf = (JSAMPLE *)malloc(cinfo.output_width * 3);
        }
    }

    if (*jpgfmt != FMT_JPEG_EXIF && cinfo.saw_JFIF_marker) {
        if (cinfo.density_unit == 1) {
            v.ippi[0] = (double)cinfo.X_density;
            v.ippi[1] = (double)cinfo.Y_density;
        } else if (cinfo.density_unit == 2) {
            v.ippi[0] = (double)cinfo.X_density * 2.54;
            v.ippi[1] = (double)cinfo.Y_density * 2.54;
        }
        if (cinfo.X_density && cinfo.Y_density) {
            if (cinfo.density_unit > 0)
                v.par = (double)cinfo.X_density / (double)cinfo.Y_density;
            else
                v.par = (double)cinfo.Y_density / (double)cinfo.X_density;
        }
        sprintf(v.xinfostr,
                "%s: %g%s%g%s",
                cinfo.density_unit ? "Resolution" : "Pixel Aspect Ratio",
                cinfo.density_unit ? v.ippi[0] : (double)cinfo.X_density,
                cinfo.density_unit ? " × " : ":",
                cinfo.density_unit ? v.ippi[1] : (double)cinfo.Y_density,
                cinfo.density_unit ? " pixels per inch" : "");
        *jpgfmt = FMT_JPEG_JFIF;
    }

    w = cinfo.output_width; h = cinfo.output_height;

    v.pad = 0;
    img = mkximage(w, h, xv.depth, &v.pad);

    row_stride = cinfo.output_width * cinfo.output_components;
    buf = (*cinfo.mem->alloc_sarray)((j_common_ptr)&cinfo, JPOOL_IMAGE,
            row_stride, 4);

    (void)jpeg_start_decompress(&cinfo);

    dp = (unsigned char *)img->data;
    idp = (unsigned int *)dp;

    while (cinfo.output_scanline < cinfo.output_height) {
        int nl = jpeg_read_scanlines(&cinfo, buf, 4);
        for (y = 0; y < nl; y++) {
            bp = buf[y];
            int nc = cinfo.output_components;
            if (v.hTransform) {
                JSAMPLE *obp = nc == 1 ? g2rbuf : bp;
                cmsDoTransform(v.hTransform, bp, obp,
                        cinfo.output_width);
                bp = obp;
                nc = 3;
            }
            if (nc >= 3) {
                rgbp = (RGB *)bp;
                for (x = 0; x < cinfo.output_width; x++) {
                    rgb = *rgbp++;
                    DOPIXEL(rgb.r, rgb.g, rgb.b, 255);
                }
            } else { 
                for (x = 0; x < cinfo.output_width; x++) {
                    DOPIXEL(*bp, *bp, *bp, 255);
                    bp++;
                }
            }
            if (v.pad) dp += v.pad;
        }
    }

    if (v.hTransform) {
        cmsDeleteTransform(v.hTransform); v.hTransform = NULL;
        if (g2rbuf) free(g2rbuf);
    }
    if (v.hInProfile) {
        if (v.iprof) cmsCloseProfile(v.hInProfile); v.hInProfile = NULL;
    }
    if (v.iprof) free(v.iprof);
    (void)jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    fclose(f);
    return img;
}
