/*
 *  ImgView  A simple image viewer
 *
 *  Copyright (C) 2013 Mats Peterson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  Please send any comments/bug reports to
 *  matsp888@yahoo.com  (Mats Peterson)
 */

#include <xsimple.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "common.h"
#include "../imgview.h"

typedef struct {
    char manufacturer;
    char version;
    char encoding;
    char bitsperpixel;
    short xmin, ymin;
    short xmax, ymax;
    short hres, vres;
    RGB palette[16];
    char reserved;
    char planes;
    short bytesperline;
    short palettetype;
    char filler[58];
} PCXHDR;

static RGB egapal[16] = {
    {0, 0, 0}, {0, 0, 170}, {0, 170, 0}, {0, 170, 170},
    {170, 0, 0}, {170, 0, 170}, {170, 85, 0}, {170, 170, 170},
    {85, 85, 85}, {85, 85, 255}, {85, 255, 85}, {85, 255, 255},
    {255, 85, 85}, {255, 85, 255}, {255, 255, 85}, {255, 255, 255}
};


static void errmsg(char *msg)
{
    message("PCX Error", msg, NULL);
}

static void planes_to_pixels(unsigned char *pixels, unsigned char *bitplanes,
	int bytesperline, int planes, int bitsperpixel)
{
    int	i, j;
    int	npixels;
    unsigned char *p;

    /* truecolor */
    if (bitsperpixel == 8) {
        unsigned char *bp = bitplanes;
        for (i = 0; i < 3; i++) {
            p = pixels + i;
            for (j = 0; j < bytesperline; j++, p += 3)
                *p = *(bp++);
        }
        return;
    }

   /* clear the pixel buffer */
    npixels = (bytesperline * 8) / bitsperpixel;
    p = pixels;
    while (--npixels >= 0)
        *p++ = 0;

   /* do the format conversion */
    for (i = 0; i < planes; i++) {
        int pixbit, bits, mask;

        p = pixels;
        pixbit = (1 << i);
        for (j = 0; j < bytesperline; j++) {
            bits = *bitplanes++;
            for (mask = 0x80; mask != 0; mask >>= 1, p++) {
                if (bits & mask)
                    *p |= pixbit;
            }
        }
    }
}

static void unpack_pixels(unsigned char *pixels, unsigned char *bitplanes,
	int bytesperline, int bitsperpixel)
{
    register int bits;

    if (bitsperpixel == 8) {
        memcpy(pixels, bitplanes, bytesperline);
    } else if (bitsperpixel == 4) {
        while (--bytesperline >= 0) {
            bits = *bitplanes++;
            *pixels++ = (bits >> 4) & 0x0f;
            *pixels++ = bits & 0x0f;
        }
    } else if (bitsperpixel == 2) {
        while (--bytesperline >= 0) {
            bits = *bitplanes++;
            *pixels++ = (bits >> 6) & 0x03;
            *pixels++ = (bits >> 4) & 0x03;
            *pixels++ = (bits >> 2) & 0x03;
            *pixels++ = bits & 0x03;
        }
    }
}

static int readline(FILE *f, unsigned char *buf, int bytes)
{
    unsigned char *bp = buf;
    int c, count;

    while (bytes > 0) {
        if ((c = getc(f)) == EOF) {
            errmsg("Premature end of file");
            return 0;
        }
        if ((c & 0xc0) != 0xc0) {
            *(bp++) = c;
            --bytes;
            continue;
        }
        count = c & 0x3f;
        if ((c = getc(f)) == EOF) {
            errmsg("Premature end of file");
            return 0;
        }
        if (count > bytes) {
            char s[128];
            sprintf(s, "Repeat count spans end of image, "
                    "count = %d, bytes = %d", count, bytes);
            errmsg(s);
            return 0;
        }
        bytes -= count;
        while (--count >= 0)
            *(bp++) = c;
    }

    return 1;
}

XImage *readpcx(int fd)
{
    XImage *img = NULL;
    FILE *f;
    PCXHDR hdr;
    int	w, h;
    RGB pal256[256], *pal = NULL;
    int truecolor = 0, bilvl = 0;
    unsigned char *line = NULL, *pixels = NULL;
    int linebytes;
    RGB rgb, *rgbp;
    int x, y;

    f = fdopen(fd, "r");

    if (fread(&hdr, 1, sizeof(PCXHDR), f) < sizeof(PCXHDR)) {
        errmsg("Premature end of file");
        goto err;
    }

    sprintf(v.xinfostr,
            "Resolution: %d × %d pixels per inch\n"
            "Bits Per Pixel Per Plane: %d\n"
            "Planes: %d\n"
            "Version: %d",
            hdr.hres, hdr.vres, hdr.bitsperpixel, hdr.planes,
            hdr.version);

    switch (hdr.bitsperpixel) {
        case 1:
            if (hdr.planes > 4)
                goto fail;
            break;
        case 2:
        case 4:
            if (hdr.planes != 1)
                goto fail;
            break;
        case 8:
            if (hdr.planes != 1) {
                if (hdr.planes == 3 || hdr.planes == 4)
                    truecolor++;
                else
                    goto fail;
            }
            break;
        default:
        fail: {
            char s[128];
            sprintf(s, "Can't handle %d bits per pixel image with "
                    "%d planes\n", hdr.bitsperpixel, hdr.planes);
            errmsg(s);
            goto err;
        }
    }

    if (hdr.bitsperpixel == 1 && hdr.planes == 1) {
        bilvl++; v.black = 0; v.white = 1;
    }

    if (! truecolor) {
        if (hdr.bitsperpixel == 8) {
            fseek(f, -(sizeof(pal256) + 1), SEEK_END);
            if (getc(f) != 0x0c) {
                errmsg("Bad color map signature");
                goto err;
            }
            fread(pal256, 1, sizeof(pal256), f);
            fseek(f, sizeof(PCXHDR), SEEK_SET);
            pal = pal256;
        } else {
            if (! bilvl)
                pal = (hdr.version == 0 || hdr.version == 3 ? egapal :
                        hdr.palette);
        }
    }

    w = (hdr.xmax - hdr.xmin) + 1; h = (hdr.ymax - hdr.ymin) + 1;

    v.pad = 0;
    img = mkximage(w, h, bilvl ? 1 : xv.depth, &v.pad);

    dp = (unsigned char *)img->data;
    idp = (unsigned int *)dp;

    linebytes = hdr.bytesperline * hdr.planes;
    line = (unsigned char *)malloc(linebytes);

    if (! bilvl)
        pixels = (unsigned char *)malloc(truecolor ? hdr.bytesperline * 3 :
                hdr.bytesperline * 8 / hdr.bitsperpixel);

    for (y = 0; y < h; y++) {
        if (! readline(f, line, linebytes)) {
            rmximage(img); img = NULL; goto err;
        }
        if (hdr.planes == 1) {
            if (! bilvl)
                unpack_pixels(pixels, line, hdr.bytesperline,
                        hdr.bitsperpixel);
        } else
            planes_to_pixels(pixels, line, hdr.bytesperline, hdr.planes,
		    hdr.bitsperpixel);
        if (bilvl) {
            memcpy(dp, line, img->bytes_per_line); dp += img->bytes_per_line;
        } else if (truecolor) {
            rgbp = (RGB *)pixels;
            for (x = 0; x < w; x++) {
                rgb = *rgbp++;
                DOPIXEL(rgb.r, rgb.g, rgb.b, 255);
            }
        } else {
            for (x = 0; x < w; x++) {
                int ndx = pixels[x];
                DOPIXEL(pal[ndx].r, pal[ndx].g, pal[ndx].b, 255);
            }
        }
        if (v.pad) dp += v.pad;
    }

err:
    if (line) free(line);
    if (pixels) free(pixels);
    fclose(f);
    return img;
}
