/*
 *  ImgView  A simple image viewer
 *
 *  Copyright (C) 2013 Mats Peterson
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *  
 *  Please send any comments/bug reports to
 *  matsp888@yahoo.com  (Mats Peterson)
 */

#include <xsimple.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <tiffio.h>
#include <png.h>
#include "common.h"
#include "../imgview.h"


static void errhandler(const char *module, const char *fmt, va_list ap)
{
    char msg[8192];
    vsnprintf(msg, 8192, fmt, ap);
    message("TIFF Library Error", msg, NULL);
}

XImage *readtiff(int fd, char *fname, int page, int *npages, int *tiffmt)
{
    XImage *img = NULL;
    TIFF *tiff;
    uint32 exififd;
    uint32 *buf = NULL, *bp;
    unsigned char *gbuf = NULL;
    uint32 w, h;
    uint16 bps, spp, photo, comp;
    uint16 nxs, *xstype;
    int bilvl = 0, gray = 0, cmyk = 0;
    RGB bg, rgb;
    int x, y;
    int i;

    TIFFSetErrorHandler(errhandler);

    if (! (tiff = TIFFFdOpen(fd, fname, "r"))) {
        close(fd);
        return NULL;
    }

    if (! *npages) {
        while (TIFFReadDirectory(tiff)) (*npages)++; (*npages)++;
        if (*npages > 1) TIFFSetDirectory(tiff, 0);
    } else if (page != 0)
        TIFFSetDirectory(tiff, page);

    *tiffmt = TIFFGetField(tiff, TIFFTAG_EXIFIFD, &exififd) ?
            FMT_TIFF_EXIF : FMT_TIFF_STD;

    TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &h);
 
    TIFFGetFieldDefaulted(tiff, TIFFTAG_BITSPERSAMPLE, &bps);
    TIFFGetFieldDefaulted(tiff, TIFFTAG_SAMPLESPERPIXEL, &spp);
    TIFFGetFieldDefaulted(tiff, TIFFTAG_EXTRASAMPLES, &nxs, &xstype);
    TIFFGetField(tiff, TIFFTAG_PHOTOMETRIC, &photo);
    TIFFGetField(tiff, TIFFTAG_COMPRESSION, &comp);

    if (spp <= 2 && (photo == PHOTOMETRIC_MINISBLACK ||
            photo == PHOTOMETRIC_MINISWHITE)) {
        if (spp == 1 && bps == 1) {
            bilvl++;
            if (photo == PHOTOMETRIC_MINISBLACK) {
                v.black = 0; v.white = 1;
            } else {
                v.black = 1; v.white = 0;
            }
        } else {
            gray++;
        }
    } else if (photo == PHOTOMETRIC_SEPARATED) {
        cmyk++;
    }

    if (! bilvl && v.hOutProfile) {
        if (TIFFGetField(tiff, TIFFTAG_ICCPROFILE,
                    (uint32 *)&v.iproflen, &v.iprof))
            v.hInProfile = cmsOpenProfileFromMem(v.iprof, v.iproflen);
        else
            v.hInProfile = gray ? v.hGrayProfile :
                    cmyk ? v.hCMYKProfile : v.hRGBProfile;
        if (v.hInProfile && !isoutprof()) {
            v.hTransform = cmsCreateTransform(
                    v.hInProfile,
                    gray ? TYPE_GRAYA_8 : cmyk ? TYPE_CMYK_8 :
                            TYPE_RGBA_8,
                    v.hOutProfile,
                    TYPE_RGBA_8,
                    INTENT_PERCEPTUAL, 0);
        }
    }

    {
        float xres, yres;
        uint16 unit = RESUNIT_INCH;
        char res_str[80] = "";
        if (TIFFGetField(tiff, TIFFTAG_XRESOLUTION, &xres) &&
                TIFFGetField(tiff, TIFFTAG_YRESOLUTION, &yres)) {
            TIFFGetField(tiff, TIFFTAG_RESOLUTIONUNIT, &unit);
            if (unit == RESUNIT_CENTIMETER) {
                v.ippi[0] = (double)xres * 2.54;
                v.ippi[1] = (double)yres * 2.54;
            } else {
                v.ippi[0] = (double)xres;
                v.ippi[1] = (double)yres;
            }
            if (xres && yres) v.par = (double)xres / (double)yres;
            sprintf(res_str, "Resolution: %g × %g%s\n",
                    v.ippi[0], v.ippi[1],
                    (unit != RESUNIT_NONE) ? " pixels per inch" : "");
        }
        sprintf(v.xinfostr, "%s"
                "Bits Per Sample: %d\n"
                "Samples Per Pixel: %d\n"
                "Photometric Interpretation: %s\n"
                "Compression Scheme: %s",
                res_str, bps, spp,
                (photo == 0) ? "White Is Zero" :
                (photo == 1) ? "Black Is Zero" :
                (photo == 2) ? "RGB" :
                (photo == 3) ? "Palette" :
                (photo == 4) ? "Transparency Mask" :
                (photo == 5) ? "CMYK" :
                (photo == 6) ? "YCbCr" :
                (photo == 8) ? "CIE L*a*b*" :
                (photo == 9) ? "ICC L*a*b*" :
                (photo == 10) ? "ITU L*a*b*" :
                (photo == 32803) ? "Color Filter Array" :
                (photo == 32844) ? "CIE Log2(L)" :
                (photo == 32845) ? "CIE Log2(L)(u',v')" :
                (photo == 34892) ? "Linear Raw" :
                "Unknown",
                (comp == 1) ? "Uncompressed" :
                (comp == 2) ? "CCITT RLE" :
                (comp == 3) ? "CCITT Group 3 Fax" :
                (comp == 4) ? "CCITT Group 4 Fax" :
                (comp == 5) ? "LZW" :
                (comp == 7) ? "JPEG" :
                (comp == 8) ? "Adobe Deflate" :
                (comp == 32766) ? "NeXT" :
                (comp == 32771) ? "CCITT RLE/W" :
                (comp == 32773) ? "PackBits" :
                (comp == 32809) ? "ThunderScan" :
                (comp == 32909) ? "Pixar Log" :
                (comp == 32946) ? "Deflate" :
                (comp == 34676) ? "SGI LogLuv" :
                (comp == 34677) ? "SGI LogLuv 24-bit" :
                "Unknown");
    }

    {
        int tag[] = {TIFFTAG_ARTIST, TIFFTAG_DOCUMENTNAME,
                TIFFTAG_IMAGEDESCRIPTION, TIFFTAG_DATETIME,
                TIFFTAG_COPYRIGHT, TIFFTAG_MAKE, TIFFTAG_MODEL,
                TIFFTAG_HOSTCOMPUTER, TIFFTAG_SOFTWARE};
        char *name[] = {"Artist", "Document Name", "Description",
                "Creation Date/Time", "Copyright", "Manufacturer",
                "Model", "Host Computer", "Software"};
        int i;
        for (i = 0; i < 9; i++) {
            char *s;
            if (TIFFGetField(tiff, tag[i], &s)) {
                v.comments[v.ncomments].key = strdup(name[i]);
                if (isutf8(s))
                    v.comments[v.ncomments].text = strdup(s);
                else
                    v.comments[v.ncomments].text = strdup(win2utf8(s));
                v.ncomments++;
            }
        }
    }

    if (! bilvl) {
        buf = (uint32 *)_TIFFmalloc(w * h * sizeof(uint32));
        if (! TIFFReadRGBAImageOriented(tiff, w, h, buf,
                ORIENTATION_TOPLEFT, 1)) goto err;
        if (gray && v.hTransform)
            gbuf = (unsigned char *)malloc(w * 2);
    }

    v.pad = 0;
    img = mkximage(w, h, bilvl ? 1 : xv.depth, &v.pad);

    if (bilvl) {
        for (y = 0; y < h; y++) {
            if (TIFFReadScanline(tiff, img->data +
                    img->bytes_per_line * y, y, 0) < 0) goto err;
        }
    } else {
        dp = (unsigned char *)img->data;
        idp = (unsigned int *)dp;

        bg.r = v.tbgcolor_rgb.red >> 8;
        bg.g = v.tbgcolor_rgb.green >> 8;
        bg.b = v.tbgcolor_rgb.blue >> 8;

        for (y = 0; y < h; y++) {
            bp = buf + (y * w);
            if (gray && v.hTransform) {
                unsigned char *p1 = (unsigned char *)bp;
                unsigned char *p2 = gbuf;
                for (i = 0; i < w; i++) {
                    *p2++ = *p1;
                    *p2++ = *(p1 + 3);
                    p1 += 4;
                }
                if (nxs && (xstype[0] == EXTRASAMPLE_ASSOCALPHA ||
                        xstype[0] == EXTRASAMPLE_UNSPECIFIED)) {
                    p1 = gbuf;
                    for (x = 0; x < w; x++) {
                        unsigned char *g = p1++;
                        unsigned char *a = p1++;
                        if (*a != 0 && *a != 255)
                            *g += 255 - *a;
                    }
                }
                cmsDoTransform(v.hTransform, gbuf, bp, w);
            } else {
                if (!cmyk && nxs &&
                        (xstype[0] == EXTRASAMPLE_ASSOCALPHA ||
                        xstype[0] == EXTRASAMPLE_UNSPECIFIED)) {
                    unsigned char *tbp = (unsigned char *)bp;
                    for (x = 0; x < w; x++) {
                        unsigned char *r = tbp++;
                        unsigned char *g = tbp++;
                        unsigned char *b = tbp++;
                        unsigned char *a = tbp++;
                        if (*a != 0 && *a != 255) {
                            *r += 255 - *a;
                            *g += 255 - *a;
                            *b += 255 - *a;
                        }
                    }
                }
                if (v.hTransform) cmsDoTransform(v.hTransform, bp, bp, w);
            }
            for (x = 0; x < w; x++) {
                unsigned char a;
                rgb.r = TIFFGetR(*bp); rgb.g = TIFFGetG(*bp);
                rgb.b = TIFFGetB(*bp); a = TIFFGetA(*bp);
                if (! cmyk) {
                    if (a == 0)
                        rgb = bg;
                    else if (a != 255) {
                        png_composite(rgb.r, rgb.r, a, bg.r);
                        png_composite(rgb.g, rgb.g, a, bg.g);
                        png_composite(rgb.b, rgb.b, a, bg.b);
                    }
                }
                DOPIXEL(rgb.r, rgb.g, rgb.b, 255);
                bp++;
            }
            if (v.pad) dp += v.pad;
        }
    }

err:
    if (v.hTransform) {
        cmsDeleteTransform(v.hTransform); v.hTransform = NULL;
        if (gbuf) free(gbuf);
    }
    if (v.hInProfile) {
        if (v.iprof) cmsCloseProfile(v.hInProfile); v.hInProfile = NULL;
    }
    if (buf) _TIFFfree(buf);
    (void)TIFFClose(tiff);
    return img;
}
